#include <stdlib.h>
#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include "xbrush.h"

extern PRGID     *prgid;        /*  $B3F<o$N#I#D(B  */
extern BMPPALA   *bmpplm;       /*  $B%S%C%H%^%C%W(B  */
extern CONSPATS  *consparts;    /*  $B%3%s%=!<%k%Q!<%D(B  */

static XShmSegmentInfo shminfo;

void  ExitShm( dsp )
Display *dsp;
{
  XShmDetach( dsp, &shminfo );
  XDestroyImage( bmpplm->canv_xim );
  shmdt( shminfo.shmaddr );
  shmctl( shminfo.shmid, IPC_RMID, 0 );
  XFreePixmap( dsp, bmpplm->drag_gid );
}

/*  XImage$B$N:n@.(B  */
XImage *CreateXImage(Display *dsp, int width, int height)
{
  Visual    *vis;
  XImage    *image;
  int       depth, scr;
 
  DEBUG_PRINT(("CreateXImage()\n"));

  scr     = DefaultScreen( dsp );    /*  $B%G%U%)%k%H%9%/%j!<%sHV9f(B  */
  depth   = DefaultDepth(dsp, scr);  /*  $B%9%/%j!<%s$N?'$N?<$5(B  */
  vis     = DefaultVisual(dsp, scr); /*  $B%G%U%)%k%H%S%8%e%"%k(B  */

  //width = bmp->bounds.c.right - bmp->bounds.c.left;
  //height = bmp->bounds.c.bottom - bmp->bounds.c.top;

  image = XCreateImage(dsp, vis, depth, ZPixmap, 0, 0, width, height,
                                         32, 0);
  image->byte_order = LSBFirst;
  image->bitmap_unit = 32;
  image->bitmap_bit_order = LSBFirst;
  //image->bits_per_pixel = bmp->pixbits;
  //image->bytes_per_line = bmp->rowbytes;
  image->bits_per_pixel = 24;  /*  $B%T%/%;%k%S%C%H?t(B  */
  image->bytes_per_line = width * 3;  /*  $B%W%l!<%s$N2#$N%P%$%H?t(B  */
  image->width  = width;
  image->height = height;
  //image->data = bmp->baseaddr;
  image->data =
             malloc(sizeof(unsigned char) * image->bytes_per_line * height);
  image->red_mask = vis->red_mask;
  image->green_mask = vis->green_mask;
  image->blue_mask = vis->blue_mask;

  DEBUG_PRINT(("width %d  height %d  bits_per_rgb %d  depth %d\n",
           image->width, image->height, vis->bits_per_rgb, depth));

  return(image);
}

/*  SHM$B%T%C%/%9%^%C%W$N:n@.(B  */
XImage *CreateShmPixmap(Pixmap *pix, BMP *bmp, Widget wid,
                                                int width, int height )
{
  int   major, minor;
  Bool  pixmaps;

  Display   *dsp;
  Window    win;
  Visual    *vis;
  XImage    *image;
  int       depth, scr;
 
  dsp = XtDisplay( wid );  /*  $B%G%#%9%W%l%$(B  */
  win = XtWindow( wid );   /*  $B%&%#%s%I%&(B  */

  scr     = DefaultScreen( dsp );    /*  $B%G%U%)%k%H%9%/%j!<%sHV9f(B  */
  depth   = DefaultDepth(dsp, scr);  /*  $B%9%/%j!<%s$N?'$N?<$5(B  */
  vis     = DefaultVisual(dsp, scr); /*  $B%G%U%)%k%H%S%8%e%"%k(B  */

  if( XShmQueryExtension(dsp) != True )
  {
    fprintf( stderr, " Dont support MIT-SHM \n" );
    exit(0);
  }

  XShmQueryVersion( dsp, &major, &minor, &pixmaps );
  if( pixmaps != True )
  {
    fprintf( stderr, " Dont support MIT-SHM (Pixmap) \n" );
    exit(0);
  }

  image = XShmCreateImage( dsp, vis, depth, ZPixmap, 0,
			    &shminfo, width, height );

  shminfo.shmid = shmget
    ( IPC_PRIVATE, image->bytes_per_line * image->height,
      IPC_CREAT | 0777 );
  shminfo.shmaddr = image->data = shmat( shminfo.shmid, 0, 0 );
  shminfo.readOnly = False;
  if ( !XShmAttach( dsp, &shminfo ) )
    fprintf( stderr, " Cannot attach MITSHM \n" );

  if(bmp != NULL)
  {
    bmp->planes = 1;
    bmp->pixbits = image->bits_per_pixel;
    bmp->rowbytes = image->bytes_per_line;
    bmp->bounds.c.left = 0;   bmp->bounds.c.top = 0;
    bmp->bounds.c.right  = image->width;
    bmp->bounds.c.bottom = image->height;
    bmp->baseaddr = image->data;
  }

  *pix = XShmCreatePixmap( dsp, win, shminfo.shmaddr, &shminfo,
			width, height, depth );
  return( image );
}

void Clr_xim(XImage *image)  /* Clear main XImage */
{
  int pixel_byte, width, height, x, y;

  width  = bmpplm->imag_bmp.bounds.c.right - bmpplm->imag_bmp.bounds.c.left;
  height = bmpplm->imag_bmp.bounds.c.bottom - bmpplm->imag_bmp.bounds.c.top;

  pixel_byte = bmpplm->imag_bmp.pixbits / 8;

  for( y = 0; y < image->height; y++ )
    for( x = 0; x < image->width; x++ )
      memset( image->data +
              ( x + y * image->width ) * (image->bits_per_pixel / 8),
	      0xA0, 3 );
}

/*  Resize Ximage  */
void Resize_xim( Widget wid, caddr_t client_data, XEvent *event )
{
  RECT rect;

  if(event->type == ConfigureNotify)
  {
    Wget_wrk(prgid->canv_wid, &rect);
                                            /*  $B%&%#%s%I%&I=<(NN0h$N0LCV(B  */
  if( (rect.c.right - rect.c.left)
                != (bmpplm->wid_rect.c.right - bmpplm->wid_rect.c.left) || 
      (rect.c.bottom - rect.c.top)
                != (bmpplm->wid_rect.c.bottom - bmpplm->wid_rect.c.top) )

    Wget_wrk(prgid->canv_wid, &bmpplm->wid_rect);
                                            /*  $B%&%#%s%I%&I=<(NN0h$N0LCV(B  */
    Chg_windsize(prgid, bmpplm, consparts);
                                /*  $BI=<(%&%#%s%I%&$NBg$-$5JQ99(B  */
  }
}
