#include <stdlib.h>
#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <setjmp.h>
#include <jpeglib.h>

typedef struct size {
  short  h;
  short  v;
} SIZE;

jmp_buf j_buf;

METHODDEF(void)my_error_exit( j_common_ptr head ) /* Error */
{
  longjmp( j_buf, 1 );
}

/*  jpeg $B%U%!%$%k$NFI$_9~$_(B  */
int LoadJpeg(Display *dsp, Visual *vis, FILE *fp, XImage *ximage,
                                      SIZE *size, int flg)
{
  struct jpeg_decompress_struct  head;
  struct jpeg_error_mgr  jpeg_err;
  JSAMPARRAY  buff_l;
  char *buffer;
  long  buf_adr, org_adr;
  int   i, line_size, pixel_byte, scr, planes, depth;
  char co[3];

  scr      = DefaultScreen(dsp);    /*  $B%G%U%)%k%H%9%/%j!<%sHV9f(B  */
  planes   = DisplayPlanes(dsp, scr); /*  $B%G%#%9%W%l%$$N%W%l!<%s?t(B  */
  depth    = DefaultDepth( dsp, scr );  /*  $B%9%/%j!<%s$N?'$N?<$5(B  */

  head.err = jpeg_std_error( &jpeg_err );
  jpeg_err.error_exit = my_error_exit;

  if( setjmp( j_buf ) )
  {
    jpeg_destroy_decompress( &head );
    return(0);
  }

  jpeg_create_decompress( &head );
  jpeg_stdio_src( &head, fp );
  jpeg_read_header( &head, TRUE );  /*  jpeg $B%X%C%@$rF@$k(B  */

  jpeg_start_decompress( &head );
  line_size = head.output_width * head.output_components;

  buff_l = ( *head.mem->alloc_sarray )( ( j_common_ptr )&head, JPOOL_IMAGE,
				      line_size, 1 );

  size->h = head.output_width; size->v = head.output_height;

  if( flg == 0 )
    return(1);  /* flg $B$,(B 0 $B$N>l9g$O!"2hA|$NBg$-$5$@$1$r<h$k(B  */

  pixel_byte = ximage->bits_per_pixel / 8;
                                         /* Get pixel bits of XImage */

  buffer = (char *)malloc( line_size );  /*  $B2hA|%G!<%?%P%C%U%!$r:n$C$F$*$/(B  */

  while( head.output_scanline < head.output_height )
  {
    jpeg_read_scanlines( &head, buff_l, 1 );
    memcpy( &(buffer[0]), buff_l[0], line_size );
    for( i = 0; i < head.output_width; i ++ )
    {
      if( ximage->byte_order != MSBFirst )
      {
	org_adr = ( ( head.output_scanline - 1 ) * ximage->bytes_per_line +
                                       i * pixel_byte);
	buf_adr = ( i + 1 ) * head.output_components;
        if( depth == 16 )  /*  $B%&%#%s%I%&$,(B 16bit $B%+%i!<$N>l9g(B  */
	{
          co[0] = (*(buffer + buf_adr - 1) & 0xff) >> 3;
          co[1] = (*(buffer + buf_adr - 2) & 0xff) >> 2;
          co[2] = (*(buffer + buf_adr - 3) & 0xff) >> 3;
          *(ximage->data + org_adr     ) = co[0] | co[1] << 5;
          *(ximage->data + org_adr + 1 ) = co[1] >> 3 | co[2] << 3;
        }
	else
        {
	  memcpy( ximage->data + org_adr    ,  buffer + buf_adr - 1, 1 );
	  memcpy( ximage->data + org_adr + 1,  buffer + buf_adr - 2, 1 );
	  memcpy( ximage->data + org_adr + 2,  buffer + buf_adr - 3, 1 );
        }
      }
      else
      {
	org_adr = ( ( head.output_scanline - 1 ) * head.output_width + i )
	            * pixel_byte;
	buf_adr = i * head.output_components;
        if( pixel_byte < head.output_components )
          memcpy( ximage->data + org_adr,
                                   buffer + buf_adr, pixel_byte );
        else
          memcpy( ximage->data + org_adr,
                                   buffer + buf_adr, head.output_components );
        if( depth == 16 )  /*  $B%&%#%s%I%&$,(B 16bit $B%+%i!<$N>l9g(B  */
	{
          co[0] = (*(ximage->data + org_adr    ) & 0xff) >> 3;
          co[1] = (*(ximage->data + org_adr + 1) & 0xff) >> 2;
          co[2] = (*(ximage->data + org_adr + 2) & 0xff) >> 3;
          *(ximage->data + org_adr     ) = co[0] | co[1] << 5;
          *(ximage->data + org_adr + 1 ) = co[1] >> 3 | co[2] << 3;
        }
      }
    }
  }
  free( buffer );

  jpeg_finish_decompress( &head );
  jpeg_destroy_decompress( &head );

  return(1);
}

/*  jpeg $B%U%!%$%k$N=q$-9~$_(B  */
void SaveJpeg(Display *dsp, Visual *vis, char *listfile,  XImage *ximage,
              int x, int y, float qual_el, int width, int height)
{
  struct jpeg_compress_struct  head;
  struct jpeg_error_mgr  jpeg_err;
  JSAMPARRAY  buff_l;
  long  buf_adr, org_adr;
  int   i, j, level, line_size, pixel_byte;
  int   scr, depth;
  char *buffer, co[3];
  FILE *fp;

  scr      = DefaultScreen(dsp);    /*  $B%G%U%)%k%H%9%/%j!<%sHV9f(B  */
  depth    = DefaultDepth( dsp, scr );  /*  $B%9%/%j!<%s$N?'$N?<$5(B  */

  head.err = jpeg_std_error( &jpeg_err );
  jpeg_err.error_exit = my_error_exit;

  if( setjmp( j_buf ) )
  {
    jpeg_destroy_compress( &head );
    fprintf( stderr, "Cannot Create %s \n", listfile );
    fclose( fp );
    return;
  }

  if( (fp = fopen( listfile, "wb" )) == NULL )
  {
    fprintf( stderr, "Cannot Open %s \n", listfile );
    return;
  }

  pixel_byte = ximage->bits_per_pixel / 8;
                                         /* Get pixel bits of XImage */
  level = qual_el * 100;

  jpeg_create_compress( &head );
  jpeg_stdio_dest( &head, fp );
  head.image_width = width;
  head.image_height = height;
  head.input_components = 3;
  head.in_color_space = JCS_RGB;
  jpeg_set_defaults( &head );
  jpeg_set_quality( &head, level, TRUE );

  jpeg_start_compress( &head, TRUE );

  line_size = width * head.input_components;
  buff_l = ( *head.mem->alloc_sarray )( ( j_common_ptr )&head, JPOOL_IMAGE,
				      line_size, 1 );
  buffer = (char *)malloc( line_size );

  for( i = 0; i < height; i ++ )
  {
    for( j = 0; j < head.image_width; j ++ )
    {
      org_adr = i * ximage->bytes_per_line + j * pixel_byte;
      buf_adr = j * head.input_components;
      co[0] = *(ximage->data + org_adr    ) & 0xff;
      co[1] = *(ximage->data + org_adr + 1) & 0xff;
      co[2] = *(ximage->data + org_adr + 2) & 0xff;
      if( depth == 16 )  /*  $B%&%#%s%I%&$,(B 16bit $B%+%i!<$N>l9g(B  */
      {
        co[0] = (co[0] & 0x1f) << 3;
        co[1] = ((co[1] & 0x07) << 5) | ((co[0] & 0xe0) >> 3);
        co[2] = co[1] & 0xf8;
      }
      if( ximage->byte_order != MSBFirst )
      {
	memcpy( buffer + buf_adr + 2, co    , 1 );
	memcpy( buffer + buf_adr + 1, co + 1, 1 );
	memcpy( buffer + buf_adr    , co + 2, 1 );
      }
      else
      {
	memcpy( buffer + buf_adr    , co    , 1 );
	memcpy( buffer + buf_adr + 1, co + 1, 1 );
	memcpy( buffer + buf_adr + 2, co + 2, 1 );
      }
    }
    memcpy( buff_l[0], &(buffer[0]), line_size );
    jpeg_write_scanlines( &head, buff_l, 1 );
  }
  free( buffer );

  jpeg_finish_compress( &head );
  jpeg_destroy_compress( &head );

  fclose( fp );
}
