#include <stdlib.h>
#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <png.h>

typedef struct size {
  short  h;
  short  v;
} SIZE;

#define PNG_BYTES_TO_CHECK 8

jmp_buf j_buf;

void my_error_exit( pp, message ) /* Error */
png_structp pp;
png_const_charp message;
{
  longjmp( j_buf, 1 );
}

/*  PNG $B%U%!%$%k$NFI$_9~$_(B  */
int LoadPng(Display *dsp, Visual *vis, FILE *fp, XImage *ximage,
                                      SIZE *size, int flg)
{
  png_structp	pp;
  png_infop	read_info, end_info;
  png_voidp     io_ptr;
  png_uint_32   png_width, png_height;
  long  buf_adr, org_adr;
  int is_png, i, j, line_size, pixel_byte, png_pixel_byte, depth;
  int bit_depth, color_type, interlace_type, compression_type, filter_type;
  int number_passes, pass, scr, planes;
  char co[3];
  unsigned char head[PNG_BYTES_TO_CHECK];
  unsigned char *buffer;
 
  scr      = DefaultScreen(dsp);    /*  $B%G%U%)%k%H%9%/%j!<%sHV9f(B  */
  planes   = DisplayPlanes(dsp, scr); /*  $B%G%#%9%W%l%$$N%W%l!<%s?t(B  */
  depth    = DefaultDepth( dsp, scr );  /*  $B%9%/%j!<%s$N?'$N?<$5(B  */

  fread( head, 1, PNG_BYTES_TO_CHECK, fp );
  is_png = png_check_sig( head, PNG_BYTES_TO_CHECK );
  if (!is_png)
  {
    return(0);
  }

  pp   = png_create_read_struct( PNG_LIBPNG_VER_STRING, io_ptr, NULL, NULL );
  if( pp == NULL )
  {
    return(0);
  }

  read_info = png_create_info_struct( pp );
  if( read_info == NULL )
  {
    return(0);
  }
  end_info = png_create_info_struct( pp );
  if( end_info == NULL )
  {
    return(0);
  }

  if( setjmp( j_buf ) )
  {
    png_destroy_read_struct( &pp, &read_info, &end_info );
    return(0);
  }

  png_init_io( pp, fp );
  png_set_sig_bytes( pp, PNG_BYTES_TO_CHECK );

  png_read_info( pp, read_info );

  png_get_IHDR( pp, read_info, &png_width, &png_height, &bit_depth,
	&color_type, &interlace_type, &compression_type, &filter_type );


  number_passes = png_set_interlace_handling( pp );

  png_read_update_info( pp, read_info );

  line_size = png_get_rowbytes( pp, read_info );

  png_pixel_byte = line_size / png_width;

  size->h = png_width; size->v = png_height;

  if( flg == 0 )
    return(1);  /* flg $B$,(B 0 $B$N>l9g$O!"2hA|$NBg$-$5$@$1$r<h$k(B  */

  pixel_byte = ximage->bits_per_pixel / 8;
                                         /* Get pixel bits of XImage */

  buffer = (char *)malloc( line_size );

  for ( pass = 0; pass < number_passes; pass ++ )
  {
    for( j = 0; j < png_height; j ++ )
    {
      png_read_rows( pp, &buffer, NULL, 1 );
      for( i = 0; i < png_width; i ++ )
      {
	org_adr = j * ximage->bytes_per_line + i * pixel_byte;
	buf_adr = i * png_pixel_byte;
	if( ximage->byte_order != MSBFirst )
	{
          if( depth == 16 )  /*  $B%&%#%s%I%&$,(B 16bit $B%+%i!<$N>l9g(B  */
	  {
            co[0] = (*(buffer + buf_adr - 1) & 0xff) >> 3;
            co[1] = (*(buffer + buf_adr - 2) & 0xff) >> 2;
            co[2] = (*(buffer + buf_adr - 3) & 0xff) >> 3;
            *(ximage->data + org_adr     ) = co[0] | co[1] << 5;
            *(ximage->data + org_adr + 1 ) = co[1] >> 3 | co[2] << 3;
          }
	  else
          {
	    memcpy( ximage->data + org_adr    , buffer + buf_adr + 2, 1 );
	    memcpy( ximage->data + org_adr + 1, buffer + buf_adr + 1, 1 );
	    memcpy( ximage->data + org_adr + 2, buffer + buf_adr,     1 );
          }
	}
	else
	{
	  if( pixel_byte < png_pixel_byte )
	    memcpy( ximage->data + org_adr, buffer + buf_adr, pixel_byte );
	  else
	    memcpy( ximage->data + org_adr, buffer + buf_adr, png_pixel_byte );
          if( depth == 16 )  /*  $B%&%#%s%I%&$,(B 16bit $B%+%i!<$N>l9g(B  */
	  {
            co[0] = (*(ximage->data + org_adr    ) & 0xff) >> 3;
            co[1] = (*(ximage->data + org_adr + 1) & 0xff) >> 2;
            co[2] = (*(ximage->data + org_adr + 2) & 0xff) >> 3;
            *(ximage->data + org_adr     ) = co[0] | co[1] << 5;
            *(ximage->data + org_adr + 1 ) = co[1] >> 3 | co[2] << 3;
          }
	}
      }
    }
  }
  free( buffer );

  png_read_end( pp, read_info );
  png_destroy_read_struct( &pp, &read_info, &end_info );

  return(1);
}

/*  PNG $B%U%!%$%k$N=q$-9~$_(B  */
void SavePng(Display *dsp, Visual *vis, char *listfile,  XImage *ximage,
              int x, int y, float qual_el, int width, int height)
{
  png_structp	pp;
  png_infop	write_info;
  png_voidp     io_ptr;
  int i, j, level, line_size, pixel_byte;
  int   scr, depth;
  unsigned char *buffer;
  char  co[3];
  long  buf_adr, org_adr;
  FILE *fp;

  scr      = DefaultScreen(dsp);    /*  $B%G%U%)%k%H%9%/%j!<%sHV9f(B  */
  depth    = DefaultDepth( dsp, scr );  /*  $B%9%/%j!<%s$N?'$N?<$5(B  */

  if( (fp = fopen( listfile, "wb" )) == NULL )
  {
    fprintf( stderr, "Cannot Open %s \n", listfile );
    return;
  }

  pp   = png_create_write_struct( PNG_LIBPNG_VER_STRING, io_ptr, NULL, NULL );
  if( pp == NULL )
  {
    fclose( fp );
    return;
  }

  write_info = png_create_info_struct( pp );
  if( write_info == NULL )
  {
    fclose( fp );
    return;
  }

  if( setjmp( j_buf ) )
  {
    fclose( fp );
    png_destroy_write_struct( &pp, &write_info );
    return;
  }

  png_init_io( pp, fp );

  level = 1 + qual_el * 9;
    if( level > 9 )
      level = 9;
  png_set_compression_level( pp, level );
  png_set_filler( pp, 0, PNG_FILLER_BEFORE );

  png_set_IHDR( pp, write_info, width, height, 8, PNG_COLOR_TYPE_RGB,
		PNG_INTERLACE_NONE, 0, 0 );

  png_write_info( pp, write_info );

  pixel_byte = ximage->bits_per_pixel / 8;
                                         /* Get pixel bits of XImage */
  line_size = width * 3;
  buffer = (char *)malloc( line_size );

  for( i = 0; i < height; i ++ )
  {
    for( j = 0; j < width; j ++ )
    {
      org_adr = i * ximage->bytes_per_line + j * pixel_byte;
      buf_adr = j * 3;
      co[0] = *(ximage->data + org_adr    ) & 0xff;
      co[1] = *(ximage->data + org_adr + 1) & 0xff;
      co[2] = *(ximage->data + org_adr + 2) & 0xff;
      if( depth == 16 )  /*  $B%&%#%s%I%&$,(B 16bit $B%+%i!<$N>l9g(B  */
      {
        co[0] = (co[0] & 0x1f) << 3;
        co[1] = ((co[1] & 0x07) << 5) | ((co[0] & 0xe0) >> 3);
        co[2] = co[1] & 0xf8;
      }
      if( ximage->byte_order != MSBFirst )
      {
	memcpy( buffer + buf_adr + 2, co    , 1 );
	memcpy( buffer + buf_adr + 1, co + 1, 1 );
	memcpy( buffer + buf_adr    , co + 2, 1 );
      }
      else
      {
	memcpy( buffer + buf_adr    , co    , 1 );
	memcpy( buffer + buf_adr + 1, co + 1, 1 );
	memcpy( buffer + buf_adr + 2, co + 2, 1 );
      }
    }
    png_write_rows( pp, &buffer, 1 );
  }
  free( buffer );

  png_write_end( pp, write_info );
  png_destroy_write_struct( &pp, &write_info );

  fclose( fp );
}
